/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user;

import static de.ozgcloud.user.User.*;

import java.util.Optional;
import java.util.stream.Stream;

import jakarta.enterprise.context.ApplicationScoped;

import org.bson.types.ObjectId;

import de.ozgcloud.common.logging.OzgCloudLogging;
import de.ozgcloud.user.common.errorhandling.ResourceNotFoundException;
import io.quarkus.mongodb.panache.PanacheMongoRepository;
import io.quarkus.panache.common.Parameters;

@ApplicationScoped
@OzgCloudLogging
class UserRepository implements PanacheMongoRepository<User> {

	static final String PARAM_NAME_SEARCH_BY = "searchBy";
	static final String PARAM_NAME_DELETED = "deleted";
	static final String PARAM_NAME_ORGANISATIONS_EINHEIT_ID = "organisationsEinheitId";
	static final String PARAM_NAME_TIMESTAMP = "timestamp";

	private static final String LIKE_OR = " like " + param(PARAM_NAME_SEARCH_BY) + " or ";
	private static final String SEARCH_QUERY = LASTNAME_FIELD + LIKE_OR +
			FIRSTNAME_FIELD + LIKE_OR +
			USERNAME_FIELD + LIKE_OR +
			FULL_NAME_FIELD + LIKE_OR +
			FULL_NAME_REVERSED_FIELD + LIKE_OR +
			EMAIL_FIELD + " like " + param(PARAM_NAME_SEARCH_BY);

	private static final String AND_DELETED = " and deleted = " + param(PARAM_NAME_DELETED);
	private static final String ORGANISATIONS_EINHEIT_ID_IN_QUERY = ORGANISATIONS_EINHEIT_IDS_FIELD + " in " + param(
			PARAM_NAME_ORGANISATIONS_EINHEIT_ID);
	private static final String AND_ORGANISATIONS_EINHEIT_ID_IN = " and " + ORGANISATIONS_EINHEIT_ID_IN_QUERY;
	static final String UPDATE_UNSYNCED_USER_QUERY = LAST_SYNC_TIMESTAMP_FIELD + " < " + param(PARAM_NAME_TIMESTAMP);

	private static String param(String paramName) {
		return String.format(":%s", paramName);
	}

	public User updateUser(User userToUpdate) {
		var user = findById(userToUpdate.getId());
		userToUpdate.setUserSettings(user.getUserSettings());
		update(userToUpdate);

		return findById(user.getId());
	}

	public Optional<User> findByExternalId(String externalId) {
		return find(EXTERNAL_ID_FIELD, externalId).firstResultOptional();
	}

	public Optional<User> findById(String id) {
		return Optional.ofNullable(id)
				.filter(ObjectId::isValid)
				.map(ObjectId::new)
				.flatMap(this::findByIdOptional);
	}

	public Optional<User> findByEmail(String email) {
		return find(EMAIL_FIELD, email).firstResultOptional();
	}

	public Stream<User> findUsers(String query, int limit) {
		var params = Parameters.with(PARAM_NAME_SEARCH_BY, toRegex(query));

		return find(SEARCH_QUERY, params).range(0, limit - 1).stream();
	}

	public Stream<User> findUsersByDeleted(String query, boolean deleted, int limit) {
		var params = Parameters.with(PARAM_NAME_SEARCH_BY, toRegex(query)).and(PARAM_NAME_DELETED, deleted);

		return find("(" + SEARCH_QUERY + ")" + AND_DELETED, params).range(0, limit - 1).stream();
	}

	public Stream<User> findUsersByDeletedAndOrganisationsEinheitId(String query, boolean deleted, String organisationsEinheitId, int limit) {
		var params = Parameters.with(PARAM_NAME_SEARCH_BY, toRegex(query))
				.and(PARAM_NAME_DELETED, deleted)
				.and(PARAM_NAME_ORGANISATIONS_EINHEIT_ID, organisationsEinheitId);

		return find("(" + SEARCH_QUERY + ")" + AND_DELETED + AND_ORGANISATIONS_EINHEIT_ID_IN, params).range(0, limit - 1).stream();
	}

	String toRegex(String query) {
		return "/.*" + query + ".*/i";
	}

	public long updateUnsyncedUsers(long lastSyncTimestamp) {
		var params = Parameters.with(PARAM_NAME_TIMESTAMP, lastSyncTimestamp);
		return update(DELETED_FIELD, true).where(UPDATE_UNSYNCED_USER_QUERY, params);
	}

	public User refresh(User user) {
		return findByIdOptional(user.getId())
				.orElse(findByExternalId(user.getExternalId()).orElseThrow(() -> new ResourceNotFoundException(User.class, user.getExternalId())));
	}

	public Optional<User> findByKeycloakId(String id) {
		return find(KEYCLOAK_USER_ID, id).firstResultOptional();
	}

	public Stream<String> findAllDeletedUserIds() {
		return find(DELETED_FIELD, true).project(UserIdProjection.class).stream().map(UserIdProjection::getAsString);
	}

	public Stream<User> findAllActiveByOrganisationsEinheitId(String organisationsEinheitId) {
		var params = Parameters.with(PARAM_NAME_ORGANISATIONS_EINHEIT_ID, organisationsEinheitId)
				.and(PARAM_NAME_DELETED, false);

		return find(ORGANISATIONS_EINHEIT_ID_IN_QUERY + AND_DELETED, params).stream();
	}

	public void deleteById(String id) {
		deleteById(new ObjectId(id));
	}

}