package de.ozgcloud.user.sync;

import static de.ozgcloud.user.User.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Arrays;
import java.util.Map;
import java.util.stream.Stream;

import de.ozgcloud.user.TestDatabaseUtils;
import jakarta.inject.Inject;

import org.bson.Document;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import com.mongodb.client.MongoClient;
import com.mongodb.client.MongoDatabase;

import de.ozgcloud.user.User;
import de.ozgcloud.user.UserService;
import de.ozgcloud.user.UserTestFactory;
import de.ozgcloud.user.common.MongoDbTestProfile;
import de.ozgcloud.user.common.lock.Lock;
import de.ozgcloud.user.keycloak.KeycloakUserRemoteService;
import io.quarkus.test.InjectMock;
import io.quarkus.test.junit.QuarkusTest;
import io.quarkus.test.junit.TestProfile;
import io.quarkus.test.junit.mockito.InjectSpy;

@QuarkusTest
@TestProfile(MongoDbTestProfile.class)
class SyncSchedulerITCase {
	private static final String USER_KEY = "User";
	private static final String LOCK_KEY = "Lock";

	@InjectSpy
	SyncScheduler scheduler;

	@Inject
	MongoClient mongoClient;

	@Inject
	UserService userService;

	@InjectMock
	KeycloakUserRemoteService keycloakUserRemoteService;

	private MongoDatabase mongoDb;

	@BeforeEach
	void init() {
		when(keycloakUserRemoteService.getAllUsers()).thenReturn(Stream.of(UserTestFactory.create()));

		mongoDb = TestDatabaseUtils.getDatabase(mongoClient);
		mongoDb.drop();

		prepareData();
	}

	private void prepareData() {
		User user1 = UserTestFactory.createBuilder().externalId("aaaaa").build();
		User user2 = UserTestFactory.createBuilder().externalId("bbbbb").build();
		persist(user1, user2);
		when(keycloakUserRemoteService.getAllUsers()).thenReturn(Stream.of(user1, user2));
	}

	private void persist(User... users) {
		Arrays.asList(users).forEach((user) -> {
			userService.save(user);
		});
	}

	@DisplayName("Test the SyncScheduler")
	@Nested
	class TestSyncScheduler {

		@Test
		void shouldSync() {
			scheduler.start();

			var syncedUsers = mongoDb.getCollection(USER_KEY).find().iterator();
			var val = syncedUsers.next().getLong(LAST_SYNC_TIMESTAMP_FIELD);
			assertThat(val).isGreaterThan(UserTestFactory.LAST_SYNC_TIMESTAMP);
		}

		@Test
		void shouldRemoveLockAfterSync() {
			scheduler.start();

			var count = mongoDb.getCollection(LOCK_KEY).countDocuments();
			assertThat(count).isZero();
		}

		@Test
		void shouldNotRemoveExistingLock() {
			insertLock(Instant.now().minus(30, ChronoUnit.MINUTES).toEpochMilli());

			scheduler.start();

			var count = mongoDb.getCollection(LOCK_KEY).countDocuments();
			assertThat(count).isOne();
		}

		@Test
		void shouldRemoveExistingLockOlderOneDay() {
			insertLock(Instant.now().minus(2, ChronoUnit.DAYS).toEpochMilli());

			scheduler.start();

			var count = mongoDb.getCollection(LOCK_KEY).countDocuments();
			assertThat(count).isZero();
		}

		@Test
		void shouldSyncAfterRemovingLock() {
			insertLock(Instant.now().minus(2, ChronoUnit.DAYS).toEpochMilli());

			scheduler.start();

			var count = mongoDb.getCollection(LOCK_KEY).countDocuments();
			assertThat(count).isZero();

			scheduler.start();

			var syncedUsers = mongoDb.getCollection(USER_KEY).find().iterator();
			var val = syncedUsers.next().getLong(LAST_SYNC_TIMESTAMP_FIELD);
			assertThat(val).isGreaterThan(UserTestFactory.LAST_SYNC_TIMESTAMP);
		}

		private void insertLock(long timestamp) {
			var lockDoc = new Document(Map.of(Lock.TIMESTAMP_FIELD, timestamp));
			mongoDb.getCollection(LOCK_KEY).insertOne(lockDoc);
		}

	}

}
